@@ -1,5 +1,5 @@
 # rawdog: RSS aggregator without delusions of grandeur.
-# Copyright 2003, 2004, 2005, 2006 Adam Sampson <ats@offog.org>
+# Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 Adam Sampson <ats@offog.org>
 #
 # rawdog is free software; you can redistribute and/or modify it
 # under the terms of that license as published by the Free Software
@@ -16,11 +16,11 @@
 # Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 # MA 02110-1301, USA, or see http://www.gnu.org/.
 
-VERSION = "2.11"
+VERSION = "2.13"
 STATE_VERSION = 2
-import feedparser, feedfinder, plugins
+import feedparser, plugins
 from persister import Persistable, Persister
-import os, time, sha, getopt, sys, re, cgi, socket, urllib2, calendar
+import os, time, getopt, sys, re, cgi, socket, urllib2, calendar
 import string, locale
 from StringIO import StringIO
 import types
@@ -31,6 +31,34 @@ try:
 except:
 	have_threading = 0
 
+try:
+	import hashlib
+except:
+	hashlib = None
+	import sha
+
+try:
+	import tidylib
+except:
+	tidylib = None
+
+try:
+	import mx.Tidy as mxtidy
+except:
+	mxtidy = None
+
+try:
+	import feedfinder
+except:
+	feedfinder = None
+
+def new_sha1(s = ""):
+	"""Return a new SHA1 hash object."""
+	if hashlib is None:
+		return sha.new(s)
+	else:
+		return hashlib.sha1(s)
+
 def set_socket_timeout(n):
 	"""Set the system socket timeout."""
 	if hasattr(socket, "setdefaulttimeout"):
@@ -40,13 +68,10 @@ def set_socket_timeout(n):
 		import timeoutsocket
 		timeoutsocket.setDefaultSocketTimeout(n)
 
+system_encoding = None
 def get_system_encoding():
 	"""Get the system encoding."""
-	try:
-		# This doesn't exist on Python 2.2.
-		return locale.getpreferredencoding()
-	except:
-		return "UTF-8"
+	return system_encoding
 
 def safe_ftime(format, t):
 	"""Format a time value into a string in the current locale (as
@@ -62,18 +87,12 @@ def format_time(secs, config):
 		format = config["timeformat"] + ", " + config["dayformat"]
 	return safe_ftime(format, t)
 
+high_char_re = re.compile(r'[^\000-\177]')
 def encode_references(s):
 	"""Encode characters in a Unicode string using HTML references."""
-	r = StringIO()
-	for c in s:
-		n = ord(c)
-		if n >= 128:
-			r.write("&#" + str(n) + ";")
-		else:
-			r.write(c)
-	v = r.getvalue()
-	r.close()
-	return v
+	def encode(m):
+		return "&#" + str(ord(m.group(0))) + ";"
+	return high_char_re.sub(encode, s)
 
 # This list of block-level elements came from the HTML 4.01 specification.
 block_level_re = re.compile(r'^\s*<(p|h1|h2|h3|h4|h5|h6|ul|ol|pre|dl|div|noscript|blockquote|form|hr|table|fieldset|address)[^a-z]', re.I)
@@ -91,6 +110,11 @@ def sanitise_html(html, baseurl, inline, config):
 	# sgmllib handles "<br/>/" as a SHORTTAG; this workaround from
 	# feedparser.
 	html = re.sub(r'(\S)/>', r'\1 />', html)
+
+	# sgmllib is fragile with broken processing instructions (e.g.
+	# "<!doctype html!>"); just remove them all.
+	html = re.sub(r'<![^>]*>', '', html)
+
 	html = feedparser._resolveRelativeURIs(html, baseurl, "UTF-8", type)
 	p = feedparser._HTMLSanitizer("UTF-8", type)
 	p.feed(html)
@@ -106,11 +130,22 @@ def sanitise_html(html, baseurl, inline, config):
 			html = "<p>" + html
 
 	if config["tidyhtml"]:
-		import mx.Tidy
-		args = { "wrap": 0, "numeric_entities": 1 }
+		args = {"numeric_entities": 1,
+		        "output_html": 1,
+		        "output_xhtml": 0,
+		        "output_xml": 0,
+		        "wrap": 0}
 		plugins.call_hook("mxtidy_args", config, args, baseurl, inline)
-		output = mx.Tidy.tidy(html, None, None,
-		                      **args)[2]
+		plugins.call_hook("tidy_args", config, args, baseurl, inline)
+		if tidylib is not None:
+			# Disable PyTidyLib's somewhat unhelpful defaults.
+			tidylib.BASE_OPTIONS = {}
+			output = tidylib.tidy_document(html, args)[0]
+		elif mxtidy is not None:
+			output = mxtidy.tidy(html, None, None, **args)[2]
+		else:
+			# No Tidy bindings installed -- do nothing.
+			output = "<body>" + html + "</body>"
 		html = output[output.find("<body>") + 6
 		              : output.rfind("</body>")].strip()
 
@@ -201,7 +236,7 @@ def string_to_html(s, config):
 	"""Convert a string to HTML."""
 	return sanitise_html(cgi.escape(s), "", True, config)
 
-template_re = re.compile(r'(__.*?__)')
+template_re = re.compile(r'(__[^_].*?__)')
 def fill_template(template, bits):
 	"""Expand a template, replacing __x__ with bits["x"], and only
 	including sections bracketed by __if_x__ .. [__else__ ..]
@@ -238,7 +273,9 @@ def fill_template(template, bits):
 					write(bits[key])
 		else:
 			write(part)
-	return f.getvalue()
+	v = f.getvalue()
+	f.close()
+	return v
 
 file_cache = {}
 def load_file(name):
@@ -262,27 +299,33 @@ def write_ascii(f, s, config):
 
 def short_hash(s):
 	"""Return a human-manipulatable 'short hash' of a string."""
-	return sha.new(s).hexdigest()[-8:]
+	return new_sha1(s).hexdigest()[-8:]
 
-def decode_structure(struct, encoding):
-	"""Walk through a structure returned by feedparser, decoding any
-	strings that haven't already been converted to Unicode."""
-	def is_dict(t):
-		return (t is dict) or (t is feedparser.FeedParserDict)
-	for (key, value) in struct.items():
-		if type(value) is str:
-			try:
-				struct[key] = value.decode(encoding)
-			except:
-				# If the encoding's invalid, at least preserve
-				# the byte stream.
-				struct[key] = value.decode("ISO-8859-1")
-		elif is_dict(type(value)):
-			decode_structure(value, encoding)
-		elif type(value) is list:
-			for item in value:
-				if is_dict(type(item)):
-					decode_structure(item, encoding)
+def ensure_unicode(value, encoding):
+	"""Convert a structure returned by feedparser into an equivalent where
+	all strings are represented as fully-decoded unicode objects."""
+
+	if isinstance(value, str):
+		try:
+			return value.decode(encoding)
+		except:
+			# If the encoding's invalid, at least preserve
+			# the byte stream.
+			return value.decode("ISO-8859-1")
+	elif isinstance(value, unicode) and type(value) is not unicode:
+		# This is a subclass of unicode (e.g.  BeautifulSoup's
+		# NavigableString, which is unpickleable in some versions of
+		# the library), so force it to be a real unicode object.
+		return unicode(value)
+	elif isinstance(value, dict):
+		d = {}
+		for (k, v) in value.items():
+			d[k] = ensure_unicode(v, encoding)
+		return d
+	elif isinstance(value, list):
+		return [ensure_unicode(v, encoding) for v in value]
+	else:
+		return value
 
 non_alphanumeric_re = re.compile(r'<[^>]*>|\&[^\;]*\;|[^a-z0-9]')
 class Feed:
@@ -305,6 +348,9 @@ class Feed:
 		else:
 			return 1
 
+	def get_state_filename(self):
+		return "feeds/%s.state" % (short_hash(self.url),)
+
 	def fetch(self, rawdog, config):
 		"""Fetch the current set of articles from the feed."""
 
@@ -339,23 +385,27 @@ class Feed:
 				handlers = handlers,
 				auth_creds = auth_creds,
 				use_im = use_im)
-		except:
-			return None
+		except Exception, e:
+			return {
+				"rawdog_exception": e,
+				"rawdog_traceback": sys.exc_info()[2],
+				}
 
-	def update(self, rawdog, now, config, p):
+	def update(self, rawdog, now, config, articles, p):
 		"""Add new articles from a feed to the collection.
 		Returns True if any articles were read, False otherwise."""
 
-		status = None
-		if p is not None:
-			status = p.get("status")
+		status = p.get("status")
 		self.last_update = now
 
 		error = None
 		non_fatal = False
 		old_url = self.url
-		if p is None:
-			error = "Error fetching or parsing feed."
+		if "rawdog_exception" in p:
+			error = "Error fetching or parsing feed: %s" % str(p["rawdog_exception"])
+			if config["showtracebacks"]:
+				from traceback import format_tb
+				error += "\n" + "".join(format_tb(p["rawdog_traceback"]))
 		elif status is None and len(p["feed"]) == 0:
 			if config["ignoretimeouts"]:
 				return False
@@ -370,7 +420,7 @@ class Feed:
 			error = "New URL:     " + p["url"] + "\n"
 			error += "The feed has moved permanently to a new URL.\n"
 			if config["changeconfig"]:
-				rawdog.change_feed_url(self.url, p["url"])
+				rawdog.change_feed_url(self.url, p["url"], config)
 				error += "The config file has been updated automatically."
 			else:
 				error += "You should update its entry in your config file."
@@ -395,10 +445,7 @@ class Feed:
 			if not non_fatal:
 				return False
 
-		decode_structure(p, p.get("encoding") or "UTF-8")
-
-		self.etag = p.get("etag")
-		self.modified = p.get("modified")
+		p = ensure_unicode(p, p.get("encoding") or "UTF-8")
 
 		# In the event that the feed hasn't changed, then both channel
 		# and feed will be empty. In this case we return 0 so that
@@ -406,9 +453,19 @@ class Feed:
 		if len(p["entries"]) == 0:
 			return False
 
+		self.etag = p.get("etag")
+		self.modified = p.get("modified")
+
 		self.feed_info = p["feed"]
 		feed = self.url
-		articles = rawdog.articles
+
+		article_ids = {}
+		if config["useids"]:
+			# Find IDs for existing articles.
+			for (hash, a) in articles.items():
+				id = a.entry_info.get("id")
+				if a.feed == feed and id is not None:
+					article_ids[id] = a
 
 		seen = {}
 		sequence = 0
@@ -421,9 +478,17 @@ class Feed:
 			seen[article.hash] = True
 			sequence += 1
 
-			if articles.has_key(article.hash):
-				articles[article.hash].update_from(article, now)
-				plugins.call_hook("article_updated", rawdog, config, article, now)
+			id = entry_info.get("id")
+			if id in article_ids:
+				existing_article = article_ids[id]
+			elif article.hash in articles:
+				existing_article = articles[article.hash]
+			else:
+				existing_article = None
+
+			if existing_article is not None:
+				existing_article.update_from(article, now)
+				plugins.call_hook("article_updated", rawdog, config, existing_article, now)
 			else:
 				articles[article.hash] = article
 				plugins.call_hook("article_added", rawdog, config, article, now)
@@ -482,13 +547,18 @@ class Article:
 			except OverflowError:
 				pass
 
-		self.hash = self.compute_hash()
+		self.hash = self.compute_initial_hash()
 
 		self.last_seen = now
 		self.added = now
 
-	def compute_hash(self):
-		h = sha.new()
+	def compute_initial_hash(self):
+		"""Compute an initial unique hash for an article.
+		The generated hash must be unique amongst all articles in the
+		system (i.e. it can't just be the article ID, because that
+		would collide if more than one feed included the same
+		article)."""
+		h = new_sha1()
 		def add_hash(s):
 			h.update(s.encode("UTF-8"))
 
@@ -508,8 +578,7 @@ class Article:
 
 	def update_from(self, new_article, now):
 		"""Update this article's contents from a newer article that's
-		been identified to be the same (i.e. has hashed the same, but
-		might have other changes that aren't part of the hash)."""
+		been identified to be the same."""
 		self.entry_info = new_article.entry_info
 		self.sequence = new_article.sequence
 		self.date = new_article.date
@@ -518,6 +587,12 @@ class Article:
 	def can_expire(self, now, config):
 		return ((now - self.last_seen) > config["expireage"])
 
+	def get_sort_date(self, config):
+		if config["sortbyfeeddate"]:
+			return self.date or self.added
+		else:
+			return self.added
+
 class DayWriter:
 	"""Utility class for writing day sections into a series of articles."""
 
@@ -586,16 +661,16 @@ def parse_list(value):
 def parse_feed_args(argparams, arglines):
 	"""Parse a list of feed arguments. Raise ConfigError if the syntax is invalid."""
 	args = {}
-	for a in argparams:
-		as = a.split("=", 1)
-		if len(as) != 2:
-			raise ConfigError("Bad feed argument in config: " + a)
-		args[as[0]] = as[1]
-	for a in arglines:
-		as = a.split(None, 1)
-		if len(as) != 2:
-			raise ConfigError("Bad argument line in config: " + a)
-		args[as[0]] = as[1]
+	for p in argparams:
+		ps = p.split("=", 1)
+		if len(ps) != 2:
+			raise ConfigError("Bad feed argument in config: " + p)
+		args[ps[0]] = ps[1]
+	for p in arglines:
+		ps = p.split(None, 1)
+		if len(ps) != 2:
+			raise ConfigError("Bad argument line in config: " + p)
+		args[ps[0]] = ps[1]
 	if "maxage" in args:
 		args["maxage"] = parse_time(args["maxage"])
 	return args
@@ -605,7 +680,8 @@ class ConfigError(Exception): pass
 class Config:
 	"""The aggregator's configuration."""
 
-	def __init__(self):
+	def __init__(self, locking):
+		self.locking = locking
 		self.files_loaded = []
 		if have_threading:
 			self.loglock = threading.Lock()
@@ -631,6 +707,7 @@ class Config:
 			"itemtemplate" : "default",
 			"verbose" : 0,
 			"ignoretimeouts" : 0,
+			"showtracebacks" : 0,
 			"daysections" : 1,
 			"timesections" : 1,
 			"blocklevelhtml" : 1,
@@ -641,6 +718,8 @@ class Config:
 			"newfeedperiod" : "3h",
 			"changeconfig": 0,
 			"numthreads": 0,
+			"splitstate": 0,
+			"useids": 0,
 			}
 
 	def __getitem__(self, key): return self.config[key]
@@ -661,7 +740,7 @@ class Config:
 		try:
 			f = open(filename, "r")
 			for line in f.xreadlines():
-				stripped = line.strip()
+				stripped = line.decode(get_system_encoding()).strip()
 				if stripped == "" or stripped[0] == "#":
 					continue
 				if line[0] in string.whitespace:
@@ -737,6 +816,8 @@ class Config:
 			self["verbose"] = parse_bool(l[1])
 		elif l[0] == "ignoretimeouts":
 			self["ignoretimeouts"] = parse_bool(l[1])
+		elif l[0] == "showtracebacks":
+			self["showtracebacks"] = parse_bool(l[1])
 		elif l[0] == "daysections":
 			self["daysections"] = parse_bool(l[1])
 		elif l[0] == "timesections":
@@ -757,6 +838,10 @@ class Config:
 			self["changeconfig"] = parse_bool(l[1])
 		elif l[0] == "numthreads":
 			self["numthreads"] = int(l[1])
+		elif l[0] == "splitstate":
+			self["splitstate"] = parse_bool(l[1])
+		elif l[0] == "useids":
+			self["useids"] = parse_bool(l[1])
 		elif l[0] == "include":
 			self.load(l[1], False)
 		elif plugins.call_hook("config_option_arglines", self, l[0], l[1], arglines):
@@ -809,7 +894,10 @@ class AddFeedEditor:
 
 def add_feed(filename, url, rawdog, config):
 	"""Try to add a feed to the config file."""
-	feeds = feedfinder.feeds(url)
+	if feedfinder is None:
+		feeds = [url]
+	else:
+		feeds = feedfinder.feeds(url)
 	if feeds == []:
 		print >>sys.stderr, "Cannot find any feeds in " + url
 	else:
@@ -916,6 +1004,12 @@ class FeedFetcher:
 		self.config.log("Thread farm finished with ", len(self.results), " results")
 		return self.results
 
+class FeedState(Persistable):
+	"""The collection of articles in a feed."""
+
+	def __init__(self):
+		self.articles = {}
+
 class Rawdog(Persistable):
 	"""The aggregator itself."""
 
@@ -924,6 +1018,7 @@ class Rawdog(Persistable):
 		self.articles = {}
 		self.plugin_storage = {}
 		self.state_version = STATE_VERSION
+		self.using_splitstate = None
 
 	def get_plugin_storage(self, plugin):
 		try:
@@ -943,7 +1038,7 @@ class Rawdog(Persistable):
 			version = 1
 		return version == STATE_VERSION
 
-	def change_feed_url(self, oldurl, newurl):
+	def change_feed_url(self, oldurl, newurl, config):
 		"""Change the URL of a feed."""
 
 		assert self.feeds.has_key(oldurl)
@@ -955,13 +1050,22 @@ class Rawdog(Persistable):
 		edit_file("config", ChangeFeedEditor(oldurl, newurl).edit)
 
 		feed = self.feeds[oldurl]
+		old_state = feed.get_state_filename()
 		feed.url = newurl
 		del self.feeds[oldurl]
 		self.feeds[newurl] = feed
 
-		for article in self.articles.values():
-			if article.feed == oldurl:
+		if config["splitstate"]:
+			persister, feedstate = load_persisted(feed.get_state_filename(), FeedState, config)
+			for article in feedstate.articles.values():
 				article.feed = newurl
+			feedstate.modified()
+			save_persisted(persister, config)
+			os.rename(old_state, feed.get_state_filename())
+		else:
+			for article in self.articles.values():
+				if article.feed == oldurl:
+					article.feed = newurl
 
 		print >>sys.stderr, "Feed URL automatically changed."
 
@@ -978,6 +1082,42 @@ class Rawdog(Persistable):
 	def sync_from_config(self, config):
 		"""Update rawdog's internal state to match the
 		configuration."""
+
+		try:
+			u = self.using_splitstate
+		except:
+			# We were last run with a version of rawdog that didn't
+			# have this variable -- so we must have a single state
+			# file.
+			u = False
+		if u is None:
+			self.using_splitstate = config["splitstate"]
+		elif u != config["splitstate"]:
+			if config["splitstate"]:
+				config.log("Converting to split state files")
+				for feed_hash, feed in self.feeds.items():
+					persister, feedstate = load_persisted(feed.get_state_filename(), FeedState, config)
+					feedstate.articles = {}
+					for article_hash, article in self.articles.items():
+						if article.feed == feed_hash:
+							feedstate.articles[article_hash] = article
+					feedstate.modified()
+					save_persisted(persister, config)
+				self.articles = {}
+			else:
+				config.log("Converting to single state file")
+				self.articles = {}
+				for feed_hash, feed in self.feeds.items():
+					persister, feedstate = load_persisted(feed.get_state_filename(), FeedState, config)
+					for article_hash, article in feedstate.articles.items():
+						self.articles[article_hash] = article
+					feedstate.articles = {}
+					feedstate.modified()
+					save_persisted(persister, config)
+					os.unlink(feed.get_state_filename())
+			self.modified()
+			self.using_splitstate = config["splitstate"]
+
 		seenfeeds = {}
 		for (url, period, args) in config["feedslist"]:
 			seenfeeds[url] = 1
@@ -1000,10 +1140,16 @@ class Rawdog(Persistable):
 		for url in self.feeds.keys():
 			if not seenfeeds.has_key(url):
 				config.log("Removing feed: ", url)
+				if config["splitstate"]:
+					try:
+						os.unlink(self.feeds[url].get_state_filename())
+					except OSError:
+						pass
+				else:
+					for key, article in self.articles.items():
+						if article.feed == url:
+							del self.articles[key]
 				del self.feeds[url]
-				for key, article in self.articles.items():
-					if article.feed == url:
-						del self.articles[key]
 				self.modified()
 
 	def update(self, config, feedurl = None):
@@ -1014,6 +1160,7 @@ class Rawdog(Persistable):
 
 		feedparser._FeedParserMixin.can_contain_relative_uris = ["url"]
 		feedparser._FeedParserMixin.can_contain_dangerous_markup = []
+		feedparser.BeautifulSoup = None
 		set_socket_timeout(config["timeout"])
 
 		if feedurl is None:
@@ -1036,43 +1183,73 @@ class Rawdog(Persistable):
 		else:
 			prefetched = {}
 
-		count = 0
 		seen_some_items = {}
+		def do_expiry(articles):
+			feedcounts = {}
+			for key, article in articles.items():
+				url = article.feed
+				feedcounts[url] = feedcounts.get(url, 0) + 1
+
+			expiry_list = []
+			feedcounts = {}
+			for key, article in articles.items():
+				url = article.feed
+				feedcounts[url] = feedcounts.get(url, 0) + 1
+				expiry_list.append((article.added, article.sequence, key, article))
+			expiry_list.sort()
+
+			count = 0
+			for date, seq, key, article in expiry_list:
+				url = article.feed
+				if url not in self.feeds:
+					config.log("Expired article for nonexistent feed: ", url)
+					count += 1
+					del articles[key]
+					continue
+				if (seen_some_items.has_key(url)
+				    and self.feeds.has_key(url)
+				    and article.can_expire(now, config)
+				    and feedcounts[url] > self.feeds[url].get_keepmin(config)):
+					plugins.call_hook("article_expired", self, config, article, now)
+					count += 1
+					feedcounts[url] -= 1
+					del articles[key]
+			config.log("Expired ", count, " articles, leaving ", len(articles))
+
+		count = 0
 		for url in update_feeds:
 			count += 1
 			config.log("Updating feed ", count, " of " , numfeeds, ": ", url)
 			feed = self.feeds[url]
+
+			if config["splitstate"]:
+				persister, feedstate = load_persisted(feed.get_state_filename(), FeedState, config)
+				articles = feedstate.articles
+			else:
+				articles = self.articles
+
 			if url in prefetched:
 				content = prefetched[url]
 			else:
 				plugins.call_hook("pre_update_feed", self, config, feed)
 				content = feed.fetch(self, config)
 			plugins.call_hook("mid_update_feed", self, config, feed, content)
-			rc = feed.update(self, now, config, content)
+			rc = feed.update(self, now, config, articles, content)
+			url = feed.url
 			plugins.call_hook("post_update_feed", self, config, feed, rc)
 			if rc:
 				seen_some_items[url] = 1
+				if config["splitstate"]:
+					feedstate.modified()
 
-		expiry_list = []
-		feedcounts = {}
-		for key, article in self.articles.items():
-			url = article.feed
-			feedcounts[url] = feedcounts.get(url, 0) + 1
-			expiry_list.append((article.added, article.sequence, key, article))
-		expiry_list.sort()
+			if config["splitstate"]:
+				do_expiry(articles)
+				save_persisted(persister, config)
 
-		count = 0
-		for date, seq, key, article in expiry_list:
-			url = article.feed
-			if (seen_some_items.has_key(url)
-			    and self.feeds.has_key(url)
-			    and article.can_expire(now, config)
-			    and feedcounts[url] > self.feeds[url].get_keepmin(config)):
-				plugins.call_hook("article_expired", self, config, article, now)
-				count += 1
-				feedcounts[url] -= 1
-				del self.articles[key]
-		config.log("Expired ", count, " articles, leaving ", len(self.articles))
+		if config["splitstate"]:
+			self.articles = {}
+		else:
+			do_expiry(self.articles)
 
 		self.modified()
 		config.log("Finished update")
@@ -1163,7 +1340,7 @@ __description__
 		itembits = {}
 		for name, value in feed.args.items():
 			if name.startswith("define_"):
-				itembits[name[7:]] = value
+				itembits[name[7:]] = sanitise_html(value, "", True, config)
 
 		title = detail_to_html(entry_info.get("title_detail"), True, config)
 
@@ -1291,9 +1468,10 @@ __description__
 <tr id="feedsheader">
 <th>Feed</th><th>RSS</th><th>Last fetched</th><th>Next fetched after</th>
 </tr>"""
-		feeds = self.feeds.values()
-		feeds.sort(lambda a, b: cmp(a.get_html_name(config).lower(), b.get_html_name(config).lower()))
-		for feed in feeds:
+		feeds = [(feed.get_html_name(config).lower(), feed)
+		         for feed in self.feeds.values()]
+		feeds.sort()
+		for (key, feed) in feeds:
 			print >>f, '<tr class="feedsrow">'
 			print >>f, '<td>' + feed.get_html_link(config) + '</td>'
 			print >>f, '<td><a class="xmlbutton" href="' + cgi.escape(feed.url) + '">XML</a></td>'
@@ -1302,6 +1480,7 @@ __description__
 			print >>f, '</tr>'
 		print >>f, """</table>"""
 		bits["feeds"] = f.getvalue()
+		f.close()
 		bits["num_feeds"] = str(len(feeds))
 
 		return bits
@@ -1323,7 +1502,8 @@ __description__
 
 		bits = self.get_main_template_bits(config)
 		bits["items"] = f.getvalue()
-		bits["num_items"] = str(len(self.articles))
+		f.close()
+		bits["num_items"] = str(len(articles))
 		plugins.call_hook("output_bits", self, config, bits)
 		s = fill_template(self.get_template(config), bits)
 		outputfile = config["outputfile"]
@@ -1342,35 +1522,53 @@ __description__
 		config.log("Starting write")
 		now = time.time()
 
-		article_dates = {}
-		articles = self.articles.values()
-		for a in articles:
-			if config["sortbyfeeddate"]:
-				article_dates[a] = a.date or a.added
-			else:
-				article_dates[a] = a.added
-		numarticles = len(articles)
+		def list_articles(articles):
+			return [(-a.get_sort_date(config), a.feed, a.sequence, a.hash) for a in articles.values()]
+		if config["splitstate"]:
+			article_list = []
+			for feed in self.feeds.values():
+				persister, feedstate = load_persisted(feed.get_state_filename(), FeedState, config)
+				article_list += list_articles(feedstate.articles)
+				save_persisted(persister, config)
+		else:
+			article_list = list_articles(self.articles)
+		numarticles = len(article_list)
 
-		def compare(a, b):
-			"""Compare two articles to decide how they
-			   should be sorted. Sort by added date, then
-			   by feed, then by sequence, then by hash."""
-			i = cmp(article_dates[b], article_dates[a])
-			if i != 0:
-				return i
-			i = cmp(a.feed, b.feed)
-			if i != 0:
-				return i
-			i = cmp(a.sequence, b.sequence)
-			if i != 0:
-				return i
-			return cmp(a.hash, b.hash)
-		plugins.call_hook("output_filter", self, config, articles)
-		articles.sort(compare)
-		plugins.call_hook("output_sort", self, config, articles)
+		if not plugins.call_hook("output_sort_articles", self, config, article_list):
+			article_list.sort()
 
 		if config["maxarticles"] != 0:
-			articles = articles[:config["maxarticles"]]
+			article_list = article_list[:config["maxarticles"]]
+
+		if config["splitstate"]:
+			wanted = {}
+			for (date, feed_url, seq, hash) in article_list:
+				if not feed_url in self.feeds:
+					# This can happen if you've managed to
+					# kill rawdog between it updating a
+					# split state file and the main state
+					# -- so just ignore the article and
+					# it'll expire eventually.
+					continue
+				wanted.setdefault(feed_url, []).append(hash)
+
+			found = {}
+			for (feed_url, article_hashes) in wanted.items():
+				feed = self.feeds[feed_url]
+				persister, feedstate = load_persisted(feed.get_state_filename(), FeedState, config)
+				for hash in article_hashes:
+					found[hash] = feedstate.articles[hash]
+				save_persisted(persister, config)
+		else:
+			found = self.articles
+
+		articles = []
+		article_dates = {}
+		for (date, feed, seq, hash) in article_list:
+			a = found.get(hash)
+			if a is not None:
+				articles.append(a)
+				article_dates[a] = -date
 
 		plugins.call_hook("output_write", self, config, articles)
 
@@ -1416,11 +1614,40 @@ Special actions (all other options are ignored if one of these is specified):
 
 Report bugs to <ats@offog.org>."""
 
+def load_persisted(fn, klass, config, no_block = False):
+	"""Attempt to load a persisted object. Returns the persister and the
+	object."""
+	config.log("Loading state file: ", fn)
+	persister = Persister(fn, klass, config.locking)
+	try:
+		obj = persister.load(no_block = no_block)
+	except KeyboardInterrupt:
+		sys.exit(1)
+	except:
+		print "An error occurred while reading state from " + os.getcwd() + "/" + fn + "."
+		print "This usually means the file is corrupt, and removing it will fix the problem."
+		sys.exit(1)
+	return (persister, obj)
+
+def save_persisted(persister, config):
+	if persister.object.is_modified():
+		config.log("Saving state file: ", persister.filename)
+	persister.save()
+
 def main(argv):
 	"""The command-line interface to the aggregator."""
 
 	locale.setlocale(locale.LC_ALL, "")
 
+	global system_encoding
+	try:
+		# This doesn't exist on Python 2.2.
+		# It's also quite expensive, which is why we do it on startup
+		# and cache the result.
+		system_encoding = locale.getpreferredencoding()
+	except:
+		system_encoding = "UTF-8"
+
 	try:
 		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:r:NW", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add=", "remove=", "no-locking", "no-lock-wait"])
 	except getopt.GetoptError, s:
@@ -1468,28 +1695,21 @@ def main(argv):
 
 	sys.path.append(".")
 
-	config = Config()
-	try:
-		config.load("config")
-	except ConfigError, err:
-		print >>sys.stderr, "In config:"
-		print >>sys.stderr, err
-		return 1
-	if verbose:
-		config["verbose"] = True
-
-	persister = Persister("state", Rawdog, locking)
-	try:
-		rawdog = persister.load(no_block = no_lock_wait)
-		if rawdog is None:
-			return 0
-	except KeyboardInterrupt:
-		return 1
-	except:
-		print "An error occurred while reading state from " + statedir + "/state."
-		print "This usually means the file is corrupt, and removing it will fix the problem."
-		return 1
+	config = Config(locking)
+	def load_config(fn):
+		try:
+			config.load(fn)
+		except ConfigError, err:
+			print >>sys.stderr, "In " + fn + ":"
+			print >>sys.stderr, err
+			return 1
+		if verbose:
+			config["verbose"] = True
+	load_config("config")
 
+	persister, rawdog = load_persisted("state", Rawdog, config, no_lock_wait)
+	if rawdog is None:
+		return 0
 	if not rawdog.check_state_version():
 		print "The state file " + statedir + "/state was created by an older"
 		print "version of rawdog, and cannot be read by this version."
@@ -1510,12 +1730,8 @@ def main(argv):
 		elif o in ("-w", "--write"):
 			rawdog.write(config)
 		elif o in ("-c", "--config"):
-			try:
-				config.load(a)
-			except ConfigError, err:
-				print >>sys.stderr, "In " + a + ":"
-				print >>sys.stderr, err
-				return 1
+			load_config(a)
+			rawdog.sync_from_config(config)
 		elif o in ("-t", "--show-template"):
 			rawdog.show_template(config)
 		elif o in ("-T", "--show-itemtemplate"):
@@ -1531,7 +1747,7 @@ def main(argv):
 
 	plugins.call_hook("shutdown", rawdog, config)
 
-	persister.save()
+	save_persisted(persister, config)
 
 	return 0
 
